
/////////////////////////////////////////////////////////////////////////////////////////////////////
//
//  BocaIpadSDK.m
//  BocaIpadSDK
//
//  Created by Michael Hall on 10/15/16.
//  Version 6.0 Introduced OpenSessionBTPicker to allow the selection of multiple Boca BT printer
//  Beta Version 6.1 is for lightening to USB - this is not complete
//  Beta Version 6.2 is to improve PDF print quality.  Quality improved, scaling needs work.
//  Beta Version 6.3 6/24/19 Added NULL tests for CFRelease which may periodically crash when passed a NULL value
//  Copyright © 2019 Boca Systems. All rights reserved.
//  Beta Version 6.4 3/24/21 Added BmpToPcx function to send an
//  RLE compressed image to the Printer. The ParseBMP function
//  was removed and replaced by the BmpToPcx function. The
//  EnableDithering function was also added.
//  Beta Version 6.5 54/7/2021 Modified
//  PcxPCX_CompressRow and PCX_CompressRowCount functions. Added SendWebImage functions.
//  Added SetStartPosition. Used to establish the printed Start Position for images.
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//  Boca Systems SDK Library Routines
//  =================================
//
//  BOOL status = OpenSessionWiFi(NSString *IP_Address)
//  void CloseSessionWiFi()
//
//  BOOL status = OpenSessionBT()
//  BOOL status = OpenSessionBTPicker(NSString *name)
//  void CloseSessionBT()
//
//  BOOL status = OpenSessionUSB()
//  void CloseSessionUSB()
//
//  void ChangeConfiguration(NSString *path, int magnifier, NSString *orientation)
//
//  void SendString(NSString *string)
//  bool SendFile(NSURL *filename, int row, int column)
//
//  void StartPosition(int row, int column)
//  bool DownloadLogo(NSURL *filename, int idnum)
//  bool PrintLogo(NSString *idnum, int row, int column)
//
//  void EnableDithering(bool bEnableDithering);
//  void ClearMemory()
//
//  void PrintCut()
//  void PrintNoCut()
//
//  NSString *PrinterStatus ReadPrinter()
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////
#include <AssetsLibrary/AssetsLibrary.h>
#include <UIKit/UIKit.h>
#import "resizeImage.h"
#import "BocaIpadSDK.h"
#import "BocaBTController.h"
#import "BocaWFController.h"
#import "BocaUSBController.h"           //@mdh 6.0
#import <Photos/Photos.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>                  //inet_addr

#define Mask8(x) ( (x) & 0xFF )

#define BP(x) ( Mask8(x) )                  //Blue
#define RP(x) ( Mask8(x >> 8 ) )            //Red
#define GP(x) ( Mask8(x >> 16) )            //Green
#define AP(x) ( Mask8(x >> 24) )            //Alpha

#define Blue(x) ( Mask8(x) )                //Blue
#define Red(x) ( Mask8(x) << 8 )            //Red
#define Green(x) ( Mask8(x) << 16)          //Green
#define Alpha(x) ( Mask8(x) << 24)          //Alpha

#define R(x) ( Mask8(x) )                   //Red
#define G(x) ( Mask8(x >> 8 ) )             //Green
#define B(x) ( Mask8(x >> 16) )             //Blue
#define A(x) ( Mask8(x >> 24) )             //Alpha

@implementation BocaIpadSDK
typedef struct
{
    char bfType0;
    char bfType1;
    uint bfSize;
    ushort bfR1;
    ushort bfR2;
    uint bfOffBits;
} FileHeader;

typedef struct
{
    uint biSize;
    uint biWidth;
    uint biHeight;
    ushort biPlanes;
    ushort biBitCount;
    uint biCompression;
    uint biSizeImage;
    uint biXPPMeter;
    uint biYPPMeter;
    uint biClrUsed;
    uint biClrImportant;
    uint biBlackBitMask;
    uint biWhiteBitMask;
} InfoHeader;

typedef struct
{
    int start;                  //starting positions for block of zeros/data
    int count;                  //number of zeros/data in block
    int block_type;             //0 for zeros and 1 for data
} LineParser;

typedef struct
{
    Byte Alpha;
    Byte Red;
    Byte Green;
    Byte Blue;
} ArgbColor;

const int LINEMAPMAX=2500;

//Stock Sizes
const int CONCERT = 0;
const int CINEMA = 1;
const int CREDITCARD = 2;
const int RECEIPT = 3;

//Monochrome Color
const int BLACK = 0;
const int WHITE = 1;
const int DITHERTHRESHOLD = 128;

NSString *CUT = @"<p>";           //FGL cut command
NSString *EJECT = @"<q>";         //FGL No cut command

bool connectionStatus;
NSString *PrinterPath = @"<P1>";                    //Default to path 1
NSString *PrinterOrientation = @"<LM>";             //Default to Landscape mode
bool flipped = false;                               //image orientation flipped //@mdh test not completed
int PrinterResolution = 300;                        //Defualt to 300 DPI
double StockHeight = 2.0;                           //Default ticket height default 2.0 inches
double StockWidth = 5.5;                            //Default ticket width default 5.5 inches
bool ImageScaled = false;
bool ImageDithered = true;
CGFloat PPI = 72;                                   //PPI - Pixels Per Inch on the display

//
NSString *Mode = @"";                               //BT for Bluetooth or WF for Wifi
NSString *s=@"";

//Barcodes and text print sharper without dithering.  Color and Grayscale images print much better with dithering
//bool Dithered = true;              //Dither or not when converting a UIImage to monochrome BMP. Default to yes.

//byte array for BMP header, compatible to Paint & FGL printer firmware
Byte BMPHeader[62];
int mDataWidth;
int nRead;
Byte *mDataArray;
Byte *mRawBitmapData;
int mDataArrayLength;
int mWidth;
int mHeight;
NSData *myImageData = nil;                                       //data read in didFinishPickingMediaWithInfo
NSString *LogoNumber = @"";

LineParser linemap[LINEMAPMAX];
FileHeader bmfile;
InfoHeader bminfo;
int portrait_dots = 0;
int portrait_dot_array[9][3];
int size_indicator = 0;
int connectedBTdevices = 0;     //V2.0 number of devices 1 to N
int selectedBTindex = 0;        //V2.0 currently selected device index 0 to N-1
int curRowPos = 10;              //Row position for image
int curColPos = 10;              //Column position for image

bool ImageToBmp(NSURL *url);
bool PdfToBmp(NSURL *url);

bool IsReadable(NSString *pathname)
{
    return [[NSFileManager defaultManager] isReadableFileAtPath:pathname];
}

int GetDirList (NSString *dir)
{
    int result = EXIT_SUCCESS;
    NSMutableSet *contents = [[NSMutableSet alloc] init];
    NSFileManager *fm = [NSFileManager defaultManager];
    BOOL isDir;
    if (dir && ([fm fileExistsAtPath:dir isDirectory:&isDir] && isDir))
    {
        printf("Recursive Directory Listing Of:\n");
        printf("%s\n\n", [dir UTF8String]);
        if (![dir hasSuffix:@"/"])
        {
            dir = [dir stringByAppendingString:@"/"];
        }
 
        // this walks the |dir| recurisively and adds the paths to the |contents| set
        NSDirectoryEnumerator *de = [fm enumeratorAtPath:dir];
        NSString *f;
        NSString *fqn;
        while ((f = [de nextObject]))
        {
            // make the filename |f| a fully qualifed filename
            fqn = [dir stringByAppendingString:f];
            if ([fm fileExistsAtPath:fqn isDirectory:&isDir] && isDir)
            {
                // append a / to the end of all directory entries
                fqn = [fqn stringByAppendingString:@"/"];
            }
            [contents addObject:fqn];
        }
 
        NSString *fn;
        // here we sort the |contents| before we display them
        for ( fn in [[contents allObjects] sortedArrayUsingSelector:@selector(compare:)] )
        {
            if (IsReadable(fn))
            {
                printf("Readable: %s\n",[fn UTF8String]);
            }
            else
            {
                printf("Not Readable: %s\n",[fn UTF8String]);
            }
        }
    }
    else
    {
        printf("%s must be directory and must exist\n", [dir UTF8String]);
        result = EXIT_FAILURE;
    }
 
    return result;
}

NSString* FindLocalFile(NSString *filename)
{
    NSFileManager *fm = [NSFileManager defaultManager];
    printf("filename = %s\n", [filename UTF8String]);
    if ([fm fileExistsAtPath:filename] == YES)
    {
        return filename;
    }
    NSString *basePath = NSHomeDirectory();
    NSString *fileFind = nil;
    if ([filename hasPrefix:@"/"])
    {
        fileFind = filename;
    }
    else
    {
        fileFind = [@"/" stringByAppendingString:filename];
    }
    NSString *testPath = [basePath stringByAppendingString:fileFind];
    printf("testPath = %s\n", [testPath UTF8String]);
    if ([fm fileExistsAtPath:testPath] == YES)
    {
        return testPath;
    }
    NSDirectoryEnumerator *de = [fm enumeratorAtPath:basePath];
    NSString *f;
    while ((f = [de nextObject]))
    {
        if (![f hasSuffix:@"/"])
        {
            f = [@"/" stringByAppendingString:f];
        }
        testPath = [basePath stringByAppendingString:f];
        printf("testPath = %s\n", [testPath UTF8String]);
        BOOL isDir = NO;
        if ([fm fileExistsAtPath:testPath isDirectory:&isDir] && isDir)
        {
            testPath = [testPath stringByAppendingString:fileFind];
            printf("testPath = %s\n", [testPath UTF8String]);
            if ([fm fileExistsAtPath:testPath] == YES)
            {
                if ([fm isReadableFileAtPath:testPath])
                {
                    printf("Path = %s\n", [testPath UTF8String]);
                    return testPath;
                }
                //printf("File Not Readable: %s\n", [testPath UTF8String]);
            }
        }
    }
    //printf("File Not Found: %s\n", [fileFind UTF8String]);
    return @"";
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
//BLUETOOTH SECTION
///////////////////////////////////////////////////////////////////////////////////////////////////////////////

BOOL OpenSessionBT()
{
    InitDotArray();                         //used for positioning in portrait mode
    connectionStatus = Open_BT();
    Mode = @"BT";                           //Set mode for Bluetooth
    
    ListUserNameAndDirectories();
    
    return(connectionStatus);
}

BOOL OpenSessionBTPicker(NSString *name)
{
    InitDotArray();                         //used for positioning in portrait
    connectionStatus = Open_BTPicker(name);
    Mode = @"BT";                           //Set mode for Bluetooth
    
    return(connectionStatus);
}

void CloseSessionBT()
{
    Close_BT();
    Mode = @"";                             //Clear mode
}


///////////////////////////////////////////////////////////////////////////////////////////////////////////////
//WI_FI SECTION
///////////////////////////////////////////////////////////////////////////////////////////////////////////////

BOOL OpenSessionWiFi(NSString *ip)
{
    InitDotArray();                         //used for positioning in portrait mode
    connectionStatus = Open_WF(ip);
    
    if(connectionStatus)
        Mode = @"WF";                       //Set mode for WiFi
    
    return(connectionStatus);
}

void CloseSessionWiFi()
{
    Close_WF();
    Mode = @"";                             //Clear mode
    connectionStatus = false;
}


///////////////////////////////////////////////////////////////////////////////////////////////////////////////
//USB/Lightning SECTION
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
//@mdh 6.0
BOOL OpenSessionUSB()
{
    InitDotArray();                         //used for positioning in portrait mode
    connectionStatus = Open_USB();
    Mode = @"USB";                           //Set mode for Bluetooth
    
    return(connectionStatus);
}

void CloseSessionUSB()
{
    Close_USB();
    Mode = @"";                             //Clear mode
}

NSURL* imagePath = nil;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
//HIGH LEVEL SUPPORT FUNCTIONS SECTION
///////////////////////////////////////////////////////////////////////////////////////////////////////////////

void SetStartPosition(int row, int column)
{
    if (CUR_POS != row)
    {
        curRowPos = row;
    }
    if (CUR_POS != column)
    {
        curColPos = column;
    }
}

//This routine will load image files into a UIImage structure, then take that image and convert it to
//a monochrome BMP file and send that to the printer.  Also raw text files can be sent directly to the
//printer through this routine
//Currently supported file types include:
// JPG, PNG and BMP - image files
// PDF - Document files, first page
// TXT - Raw text files containing FGL commands and text to be printed
//
bool SendFile(NSURL *filename, int row, int column)
{
    if (nil == filename)
    {
        printf("Error, filename is NULL\n");
        return false;
    }
    imagePath = filename;
    bool status = true;
    NSString *FileExtension = [filename.pathExtension lowercaseString];
    LogoNumber = @"";
    if (CUR_POS != row)
    {
        curRowPos = row;
    }
    if (CUR_POS != column)
    {
        curColPos = column;
    }
        
    //printf("SendFile: File = %s\n", filename.absoluteString.UTF8String);
    
    if([FileExtension isEqual: @"txt"])
    {
        status = RWTextFile(filename);
    }
    else if ([FileExtension isEqual: @"pdf"])
    {
        status = PdfToBmp(filename);
    }
    else if ([FileExtension isEqual: @"png"])
    {
        status = ImageToBmp(filename);
    }
    else if ([FileExtension isEqual: @"jpeg"])
    {
        status = ImageToBmp(filename);
    }
    else if ([FileExtension isEqual: @"bmp"])
    {
        status = ImageToBmp(filename);
    }
    else
        status = false;
    
    imagePath = nil;
    return(status);
}

bool DownloadLogo(NSURL *filename, int idnum)
{
    imagePath = filename;
    bool status = true;
    NSString *FileExtension = filename.pathExtension;
    LogoNumber = [NSString stringWithFormat:@"%d",idnum];
    
    if ([FileExtension isEqual: @"PDF"] || [FileExtension isEqual: @"pdf"])
    {
        status = PdfToBmp(filename);
    }
    else if ([FileExtension isEqual: @"PNG"] || [FileExtension isEqual: @"png"])
    {
        status = ImageToBmp(filename);
    }
    else if ([FileExtension isEqual: @"JPG"] || [FileExtension isEqual: @"jpg"] || [FileExtension isEqual: @"JPEG"] || [FileExtension isEqual: @"jpeg"])
    {
        status = ImageToBmp(filename);
    }
    else if ([FileExtension isEqual: @"BMP"] || [FileExtension isEqual: @"bmp"])
    {
        status = ImageToBmp(filename);
    }
    else
        status = false;
    
    imagePath = nil;
    return(status);
}

void EnableDithering(bool bEnableDithering)
{
    ImageDithered = bEnableDithering;
}

bool PrintLogo(int idnum, int row, int column)
{
    
    bool status = true;
    NSString *outgoing = @"<SP";
    outgoing = [NSString stringWithFormat:@"%@%d",outgoing,row];
    outgoing = [NSString stringWithFormat:@"%@%@",outgoing,@","];
    outgoing = [NSString stringWithFormat:@"%@%d",outgoing,column];
    outgoing = [NSString stringWithFormat:@"%@%@",outgoing,@"><LD"];
    outgoing = [NSString stringWithFormat:@"%@%d",outgoing,idnum];
    outgoing = [NSString stringWithFormat:@"%@%@",outgoing,@">"];
    SendString(outgoing);
    
    return(status);
}

void ChangeConfiguration(NSString *path, int magnifier, NSString *orientation)
{
    PrinterPath = path;                             //Printer Path 1, 2, 3 or 4
    PrinterResolution = magnifier;
    PrinterOrientation = orientation;               //Landscape or Portrait
    
    //str_size = _sizeOut.text;
    set_portrait_dots();
}

void PrintCut()
{
    SendString(CUT);
    
    printf("ReadPrinter = %s\n", [ReadPrinter() UTF8String]);
}

void PrintNoCut()
{
    SendString(EJECT);
}

void SendString(NSString *string)
{
    //NSLog(@"SendString Mode = %@",Mode);
    
    //if communicating via Wi-Fi or Bluetooth
    if([Mode isEqual: @"WF"])               //Wi-Fi
    {
        Write_WF(string);
    }
    else if([Mode isEqual: @"BT"])          //BlueTooth
    {
        Write_BT(string);
    }
    else if([Mode isEqual: @"USB"])          //@mdh 6.0 USB
    {
        Write_USB(string);
    }
    
}

void SendData(uint8_t *buf, int size)
{
    //NSLog(@"SendData Mode = %@",Mode);
    //printf("Begin SendData: buf = %p, size = %d\n", buf, size);
    //if communicating via Wi-Fi or Bluetooth
    if([Mode isEqual: @"WF"])
    {
        Write_WF_Data(buf, size);
    }
    else if([Mode isEqual: @"BT"])          //@mdh 6.0 BlueTooth
    {
        Write_BT_Data(buf, size);
    }
    else if([Mode isEqual: @"USB"])          //@mdh 6.0 USB
    {
        Write_USB_Data(buf, size);
    }
    //printf("End SendData: buf = %p, size = %d\n", buf, size);
}

bool SendWebImage(NSString* UrlString, int row, int column)
{
    if (CUR_POS != row)
    {
        curRowPos = row;
    }
    if (CUR_POS != column)
    {
        curColPos = column;
    }
    int BMP_WIDTH;
    int BMP_HEIGHT;
    
    if (0 == UrlString.length)
    {
        printf("Error, UrlString is NULL\n");
        return false;
    }
    NSURL *url = [NSURL URLWithString:UrlString];
    if (url && url.scheme && url.host)
    {
        NSURL *url = [NSURL URLWithString: UrlString];
        NSData *urlData = [NSData dataWithContentsOfURL:url];
        if (urlData)
        {
            UIImage *theImage = [UIImage imageWithData:urlData];
            if (theImage)
            {
                BMP_WIDTH = (int)((theImage.size.width/PPI) * PrinterResolution);
                BMP_HEIGHT = (int)((theImage.size.height/PPI) * PrinterResolution);
                PrintBMP(BMP_WIDTH, BMP_HEIGHT, theImage);
            }
            else
            {
                printf("UIImage is NULL: %s\n", [UrlString UTF8String]);
                return false;
            }
        }
        else
        {
            printf("Error, Could not load: %s\n", [UrlString UTF8String]);
            return false;
        }
    }
    else
    {
        NSString *fullPath = FindLocalFile(UrlString);
        if (0 == fullPath.length)
        {
            printf("Error, filePath is not found\n");
            return false;
        }
        UIImage *theImage = [UIImage imageWithContentsOfFile:fullPath];
        BMP_WIDTH = (int)((theImage.size.width/PPI) * PrinterResolution);
        BMP_HEIGHT = (int)((theImage.size.height/PPI) * PrinterResolution);
        PrintBMP(BMP_WIDTH, BMP_HEIGHT, theImage);
    }
    return true;
}

NSString *ReadPrinter()
{
    NSString *stat=@"";
    
    //if communicating via Wi-Fi or Bluetooth
    if([Mode isEqual: @"WF"])               //Wi-Fi
    {
        stat = Read_WF();
        ClearStatus_WF();
    }
    else if([Mode isEqual: @"BT"])          //BlueTooth
    {
        stat = Read_BT();
        ClearStatus_BT();
    }
    else if([Mode isEqual: @"USB"])          //@mdh 6.0 USB
    {
        stat = Read_USB();
        ClearStatus_USB();
    }
    return(stat);
}



///////////////////////////////////////////////////////////////////////////////////////////////////////////////
//LOW LEVEL SUPPORT FUNCTIONS SECTION
///////////////////////////////////////////////////////////////////////////////////////////////////////////////

static bool RWTextFile(NSURL *url)
{
    NSString *testFile;     //= url.absoluteString;      //[url absoluteString];
    testFile = [url path];
    bool status = true;
    bool fileExists = [[NSFileManager defaultManager] fileExistsAtPath:testFile];   //verify file existance
    
    //if file found
    if(fileExists)
    {
        NSData *myTextData = [NSData dataWithContentsOfFile:testFile];                  //read file raw data
        NSString *str=[[NSString alloc] initWithBytes:myTextData.bytes length:myTextData.length encoding:NSUTF8StringEncoding];
        //NSLog(str);
        
        //if file read
        if (myTextData)
        {
            //output file to printer as one long string
            SendString(str);
        }
        
    }
    else
    {
        status = false;
    }
    return(status);
}

- (void) resizeImage
{
    
    UIImage *newImage;
    UIImage *imageToResize;
    
    //call resize image class
    resizeImage *imageResize = [[resizeImage alloc]init];
    [imageResize resizeImage:imageToResize width:320 height:200];
    NSData *resizedImageData = [imageResize thumbnailImageData];
    
    //NSUInteger resizedImageDataSize = resizedImageData.length;
    
    //newImage = [[UIImage alloc] initWithCGImage: image.CGImage
    //                                      scale: 1.0
    //                                orientation: UIImageOrientationUp];
    
    
    newImage = [UIImage imageWithData:resizedImageData];
    //NSLog(@"Resized size %d Bytes = ", resizedImageDataSize);
}

//lg02 - removed adjustPrint routine
//1. Convert PDF file to UIImage
//2. Adjust orientation and scale
//3. Send BMP data to the printer
bool PdfToBmp(NSURL *url)
{
    int BMP_WIDTH;
    int BMP_HEIGHT;
    UIImage *image;
    CGFloat dpi = PrinterResolution/72.0;
    
    bool status = true;
    size_t PAGE_NO = 1;
    
    CGPDFDocumentRef pdf = CGPDFDocumentCreateWithURL((CFURLRef)url);
    CGPDFPageRef page = CGPDFDocumentGetPage(pdf, PAGE_NO);
    
    CGRect rect = CGPDFPageGetBoxRect(page, kCGPDFMediaBox);
    
    rect.size.width = (rect.size.width * dpi);              //V6.2
    rect.size.height = (rect.size.height * dpi);            //V6.2
    
    UIGraphicsBeginImageContext(rect.size);
    
    CGContextRef context = UIGraphicsGetCurrentContext();
    CGContextSetInterpolationQuality(context, kCGInterpolationHigh);    //V6.2
    CGContextTranslateCTM(context, 0.0, rect.size.height);              //V6.2
    //CGContextTranslateCTM(context, 0, rect.size.width);
    CGContextScaleCTM(context, dpi, -dpi);                              //V6.2
    CGContextSetGrayFillColor(context, 1.0, 1.0);
    CGContextFillRect(context, rect);
    CGContextSaveGState(context);
    CGContextDrawPDFPage(context, page);
    
    CGContextRestoreGState(context);
    
    image = UIGraphicsGetImageFromCurrentImageContext();
    
    UIImage *result;
    
    //Adjust size and orientation of image to print
    //result = AdjustPrint(image);//@mdh test V6.2
    
    result = image;
    
    //Convert pixels per inch size to Dots per inch size
    BMP_WIDTH = (int)(result.size.width * dpi);
    BMP_HEIGHT = (int)(result.size.height * dpi);
    
    //NSLog(@"PPI Width  = %f",result.size.width);
    //NSLog(@"PPI Height = %f",result.size.height);
    //NSLog(@"DPI Width  = %d",BMP_WIDTH);
    //NSLog(@"DPI Height = %d",BMP_HEIGHT);
    
    //Print the resulting image
    PrintBMP(BMP_WIDTH,BMP_HEIGHT,result);
    
    return(status);
}

NSString *FileName;

//1. Convert JPG, JPEG, PNG or BMP file to UIImage
//2. Dither into Monochrome BMP
//3. Send BMP data to the printer
bool ImageToBmp(NSURL *url)
{
    
    FileName = url.absoluteString;
    UIImage *image;
    bool status = false;
    NSDictionary *info;
    //printf("ImageToBmp: File = %s\n", FileName.UTF8String);
    
    image = [info objectForKey:@"UIImagePickerControllerOriginalImage"];
    ALAssetsLibrary *library = [[ALAssetsLibrary alloc] init];
    // PHPhotoLibrary *library = [[PHPhotoLibrary alloc] init];
    
    dispatch_semaphore_t sema = dispatch_semaphore_create(0);
    dispatch_queue_t queue = dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT,0);
    
    dispatch_async(queue, ^{
        [library assetForURL:url resultBlock:^( ALAsset *asset)
         {
            @try
            {
                ALAssetRepresentation *representation = [asset defaultRepresentation];
                
                uint8_t *buffer = (uint8_t *)malloc(sizeof(uint8_t)*(unsigned long)[representation size]);
                if (buffer != NULL)
                {
                    NSError *error = nil;
                    NSUInteger bytesRead = [representation getBytes:buffer fromOffset:0 length:(NSUInteger)[representation size] error:&error];
                    myImageData = [NSData dataWithBytes:buffer length:bytesRead];            //read in image file
                    free(buffer);
                    
                    UIImage *image=[UIImage imageWithData:myImageData];
                    UIImage *result;
                    
                    //Adjust size and orientation of image to print
                    //result = AdjustPrint(image);//@mdh test V6.2
                    
                    result = image;
                    
                    int BMP_WIDTH;
                    int BMP_HEIGHT;
                    
                    //Convert pixels per inch size to Dots per inch size
                    BMP_WIDTH = (int)((result.size.width/PPI) * PrinterResolution);
                    BMP_HEIGHT = (int)((result.size.height/PPI) * PrinterResolution);
                    
                    //NSLog(@"PPI Width  = %f",result.size.width);
                    //NSLog(@"PPI Height = %f",result.size.height);
                    //NSLog(@"DPI Width  = %d",BMP_WIDTH);
                    //NSLog(@"DPI Height = %d",BMP_HEIGHT);
                    
                    //Print the resulting image
                    PrintBMP(BMP_WIDTH,BMP_HEIGHT,result);
                    
                    dispatch_semaphore_signal(sema);
                }
                else
                {
                    printf("Error, buffer is NULL\n");
                }
            }
            @catch (NSException *exception)
            {
                NSLog(@"Exception: %@", exception.reason);
            }
            
        } failureBlock:^(NSError *error) {
            dispatch_semaphore_signal(sema);
        }];
        
    });
    dispatch_semaphore_wait(sema, DISPATCH_TIME_FOREVER);
    //dispatch_release(sema);
    
    return(status);
}

void PrintBMP(int width, int height, UIImage *image)
{
    UIImage *tempImage = nil;
    //printf("PrintBMP: Width = %f, Height = %f\n", image.size.width, image.size.height);
    
    tempImage = image;
    
    //Should we dither the image or not? Dithered images are clearer and typically print faster, due to less black.
    //If not, image may not be as clear, but barcodes print great.
    if(ImageDithered)
        CreateTransformedImage(tempImage);
    else
        ConvertToMonochrome(tempImage);
    
    createRawMonochromeData();
    populate_bmpheader();
    
    Byte *Test;
    Test = (Byte *)malloc(nRead + 62);
    //printf("PrintBMP: Test = %p, Size = %d\n", Test, nRead + 62);
    for (int i = 0; i < 62; i++)
    {
        Test[i] = BMPHeader[i];                 //move the BMP header information to the byte array
    }
    for (int i = 0; i < nRead; i++)
    {
        Test[i+62] = mRawBitmapData[i];         //move the monochrome BMP graphical data to the same byte array
    }
    
    //now that the data is housed to look like a monochrome BMP file, convert it to NSData and call BmpToPcx
    myImageData = [NSData dataWithBytes:Test length:nRead+62];
    
    BmpToPcx(myImageData);
    
    //printf("PrintBMP: Test = %p\n", Test);
    if (NULL != Test)
    {
        free(Test);
    }
}

void createRawMonochromeData()
{
    UInt32 length = 0;
    for (UInt32 i = 0; i < mDataArrayLength; i = i + 8)
    {
        Byte first = mDataArray[i];
        for (UInt32 j = 0; j < 8; j++)  //lg001 - was 7
        {
            Byte second = ((first << 1) | mDataArray[i + j]);
            first = second;
        }
        mRawBitmapData[length] = first;
        length++;
    }
    nRead = length;
}

void populate_bmpheader()
{
    BMPHeader[0] = 0x42;                            //'B'
    BMPHeader[1] = 0x4d;                            //'M'
    BMPHeader_Int((nRead + 0x3e), 2);               //Size of BMP file in bytes
    BMPHeader[6] = 0x00;                            //Reserved
    BMPHeader[7] = 0x00;
    BMPHeader[8] = 0x00;
    BMPHeader[9] = 0x00;
    BMPHeader[10] = 0x3e;                           //Offset to data
    BMPHeader[11] = 0x00;
    BMPHeader[12] = 0x00;
    BMPHeader[13] = 0x00;
    BMPHeader[14] = 0x28;                           //Size of BITMAPINFOHEADER structure, must be 40
    BMPHeader[15] = 0x00;
    BMPHeader[16] = 0x00;
    BMPHeader[17] = 0x00;
    BMPHeader_Int(mWidth, 18);                      //image width
    BMPHeader_Int(mHeight, 22);                     //image height
    BMPHeader[26] = 0x01;                           //number of planes in image monochrome = 1
    BMPHeader[27] = 0x00;
    BMPHeader[28] = 0x01;                           //number of bits per pixel monochrome = 1
    BMPHeader[29] = 0x00;
    BMPHeader[30] = 0x00;                           //compression type none = 0
    BMPHeader[31] = 0x00;
    BMPHeader[32] = 0x00;
    BMPHeader[33] = 0x00;
    BMPHeader_Int(nRead, 34);                       //image size including padding
    BMPHeader[38] = 0x00;                           //horizontal resolution in pixels per meter
    BMPHeader[39] = 0x00;
    BMPHeader[40] = 0x00;
    BMPHeader[41] = 0x00;
    BMPHeader[42] = 0x00;                           //vertical resolution in pixels per meter
    BMPHeader[43] = 0x00;
    BMPHeader[44] = 0x00;
    BMPHeader[45] = 0x00;
    BMPHeader[46] = 0x00;                           //number of colors in image, or zero
    BMPHeader[47] = 0x00;
    BMPHeader[48] = 0x00;
    BMPHeader[49] = 0x00;
    BMPHeader[50] = 0x00;                           //number of important colors, or zero
    BMPHeader[51] = 0x00;
    BMPHeader[52] = 0x00;
    BMPHeader[53] = 0x00;
    BMPHeader[54] = 0x00;                           //Padded for paint offset 0x3E
    BMPHeader[55] = 0x00;
    BMPHeader[56] = 0x00;
    BMPHeader[57] = 0x00;
    BMPHeader[58] = 0xff;                           //Padded for paint offset 0x3E
    BMPHeader[59] = 0xff;
    BMPHeader[60] = 0xff;
    BMPHeader[61] = 0x00;
    
}

//Reverse integer and load into byte array
void BMPHeader_Int(int intToReverse, int index)
{
    Byte intAsBytes[4];
    
    intAsBytes[0] = (Byte) (intToReverse & 0xFF);
    intAsBytes[1] = (Byte) ((intToReverse >> 8) & 0xFF);
    intAsBytes[2] = (Byte) ((intToReverse >> 16) & 0xFF);
    intAsBytes[3] = (Byte) ((intToReverse >> 24) & 0xFF);
    
    BMPHeader[index] = (Byte) intAsBytes[0];
    BMPHeader[index + 1] = (Byte) intAsBytes[1];
    BMPHeader[index + 2] = (Byte) intAsBytes[2];
    BMPHeader[index + 3] = (Byte) intAsBytes[3];
    
}

void ConvertToMonochrome(UIImage *image)
{
    
    int Monochrome,k;
    
    // 1.
    CGImageRef inputCGImage = [image CGImage];
    NSUInteger width = CGImageGetWidth(inputCGImage);
    NSUInteger height = CGImageGetHeight(inputCGImage);
    mWidth = (int)width;
    mHeight = (int)height;
    k=0;
    
    // 2.
    NSUInteger bytesPerPixel = 4;
    NSUInteger bytesPerRow = bytesPerPixel * width;
    NSUInteger bitsPerComponent = 8;
    
    UInt32 * pixels;
    pixels = (UInt32 *) calloc(mHeight * mWidth, sizeof(UInt32));
    
    // 3.
    CGColorSpaceRef colorSpace = CGColorSpaceCreateDeviceRGB();
    CGContextRef context = CGBitmapContextCreate(pixels, mWidth, mHeight, bitsPerComponent, bytesPerRow, colorSpace, kCGImageAlphaPremultipliedLast | kCGBitmapByteOrder32Big);
    
    // 4.
    CGContextDrawImage(context, CGRectMake(0, 0, mWidth, mHeight), inputCGImage);
    
    // 5. Cleanup
    CGColorSpaceRelease(colorSpace);
    CGContextRelease(context);
    
    // 2.
    mDataWidth = ((mWidth + 31) / 32) * 4 * 8;
    mDataArrayLength = mDataWidth * mHeight;
    mDataArray = (Byte *)malloc(mDataArrayLength);
    mRawBitmapData = (Byte *)malloc((mDataWidth * mHeight)/8);
    
    UInt32 * currentPixel = pixels;
    for (NSUInteger j = mHeight-0; j>0; j--) //lg01 - was -1
    {
        
        for (NSUInteger i = 0; i < mWidth; i++)
        {
            // 3.
            //process each row of data starting with the last row
            UInt32 color = *(currentPixel + ((j-1)*mWidth) + i);  //lg01 - was j
            Monochrome = (R(color)*0.299) + (G(color)*0.587) + (B(color)*0.114);
            
            // set new pixel color to output bitmap
            if (Monochrome < DITHERTHRESHOLD)
            {
                mDataArray[k++] = BLACK;
            }
            else
            {
                mDataArray[k++] = WHITE;
            }
        }
        
        if (mDataWidth > mWidth)
        {
            for (int p = mWidth; p < mDataWidth; p++)
            {
                mDataArray[k++] = WHITE;
            }
        }
    }
}

//Globals from rastertohp example
unsigned char *Pixels;                  //Input buffers
unsigned char *Planes;                  //Output buffers
unsigned char *CompBuffer;              //Compression buffer
unsigned char *BitBuffer;               //Buffer for output bits
UInt32 k = 0;

//Convert image to grayscale ARGB bitmap, dither and then to monochrome bitmap
void CreateTransformedImage(UIImage *image)
{
    UInt32 row,col;
    UInt32 index;
    ArgbColor current, transformed;
    ArgbColor *OriginalData;
    
    CGImageRef inputCGImage = [image CGImage];
    NSUInteger width = CGImageGetWidth(inputCGImage);
    NSUInteger height = CGImageGetHeight(inputCGImage);
    
    NSUInteger bytesPerPixel = 4;
    NSUInteger bytesPerRow = bytesPerPixel * width;
    NSUInteger bitsPerComponent = 8;
    
    //Original Data is in integer form
    UInt32 *RawData;
    RawData = (UInt32 *) calloc(height * width, sizeof(UInt32));
    //Transformed Data is in RGBA form
    OriginalData = (ArgbColor *) calloc(height * width, sizeof(ArgbColor));
    
    CGColorSpaceRef colorSpace = CGColorSpaceCreateDeviceRGB();
    
    CGContextRef context = CGBitmapContextCreate(RawData, width, height, bitsPerComponent, bytesPerRow, colorSpace, kCGImageAlphaPremultipliedLast | kCGBitmapByteOrderDefault);
    //kCGBitmapByteOrder32Big);
    
    CGContextDrawImage(context, CGRectMake(0, 0, width, height), inputCGImage);
    
    CGColorSpaceRelease(colorSpace);
    CGContextRelease(context);
    
    //Transform integer pixel data to RGBA data
    TransformToRGBA(RawData, OriginalData, (int)width, (int)height);
    
    for (row = 0; row < height; row++)
    {
        
        for (col = 0; col < width; col++)
        {
            index = ((row) * width) + col;
            //Transform the pixel to monochrome
            current = OriginalData[index];
            transformed = TransformPixel(current);
            OriginalData[index] = transformed;
            
            //Apply Floyd Steinberg dithering alogorithm
            DitherFloydSteinberg(OriginalData,current,transformed,col,row,(int)width,(int)height);
        }
    }
    
    //Return data to bitmap format
    mWidth = (int)width;
    mHeight = (int)height;
    mDataWidth = ((mWidth + 31) / 32) * 4 * 8;
    mDataArrayLength = mDataWidth * mHeight;
    mDataArray = (Byte *)malloc(mDataArrayLength);
    mRawBitmapData = (Byte *)malloc((mDataWidth * mHeight)/8);
    //Put Monochrome data into mDataArray
    MakeItMonochrome(OriginalData,(int)width,(int)height);
    
}

//Convert ARGB format to monochrome bitmap format of 0 or 1
void MakeItMonochrome(ArgbColor *originalData, int width, int height)
{
    ArgbColor RGBA;
    UInt32 row,col;
    UInt32 index;
    Byte Monochrome;
    k=0;
    
    //Transform RGBA pixel data to monochrome data
    //for (row = 0; row < height; row++)
    for (row = height; row > 0; row--)
    {
        
        for (col = 0; col < width; col++)
        {
            index = ((row-1) * width) + (col);
            RGBA = originalData[index];
            
            Monochrome = (RGBA.Red*0.299) + (RGBA.Green*0.587) + (RGBA.Blue*0.114);
            if (Monochrome < DITHERTHRESHOLD)
            {
                mDataArray[k++] = BLACK;        //Black = 0
            }
            else
            {
                mDataArray[k++] = WHITE;        //White = 1
            }
        }
        if (mDataWidth > mWidth)
        {
            for (int p = mWidth; p < mDataWidth; p++)
            {
                mDataArray[k++] = WHITE;        //White = 1
            }
        }
        
    }
    
}



//Change integer raw data to an ARGB format
ArgbColor TransformToRGBA(UInt32 *RawData, ArgbColor *OriginalData, int width, int height)
{
    ArgbColor RGBA;
    UInt32 row,col;
    UInt32 index;
    
    RGBA.Alpha = 0;
    RGBA.Red = 0;
    RGBA.Green = 0;
    RGBA.Blue = 0;
    
    //Transform integer pixel array of data to RGBA data
    for (row = 0; row < height; row++)
    {
        
        for (col = 0; col < width; col++)
        {
            index = ((row) * width) + col;
            if(index == 80)
                RGBA.Alpha = AP(RawData[index]);
            else
                RGBA.Alpha = AP(RawData[index]);
            RGBA.Red = RP(RawData[index]);
            RGBA.Green = GP(RawData[index]);
            RGBA.Blue = BP(RawData[index]);
            OriginalData[index] = RGBA;
        }
    }
    
    return(RGBA);
}



//Grayscale ARGB
ArgbColor TransformPixel(ArgbColor pixel)
{
    Byte Monochrome;
    ArgbColor NewColor;
    
    Monochrome = (pixel.Red*0.299) + (pixel.Green*0.587) + (pixel.Blue*0.114);
    
    if (Monochrome < DITHERTHRESHOLD)
    {
        //Black
        NewColor.Alpha = pixel.Alpha;
        NewColor.Red = 0;
        NewColor.Green = 0;
        NewColor.Blue = 0;
    }
    else
    {
        //White
        NewColor.Alpha = pixel.Alpha;
        NewColor.Red = 255;
        NewColor.Green = 255;
        NewColor.Blue = 255;
    }
    return (NewColor);
    
}

//Limit all values to the range 0 to 255
Byte ToByte(int value)
{
    if (value < 0)
        value = 0;
    else if (value > 255)
        value = 255;
    return (Byte)value;
}


void DitherFloydSteinberg(ArgbColor *original, ArgbColor originalPixel, ArgbColor transformedPixel, int x, int y, int width, int height)
{
    ArgbColor offsetPixel;
    
    int redError;
    int blueError;
    int greenError;
    UInt32 offsetIndex;
    UInt32 index;
    
    index = y * width + x;
    redError = originalPixel.Red - transformedPixel.Red;
    blueError = originalPixel.Green - transformedPixel.Green;
    greenError = originalPixel.Blue - transformedPixel.Blue;
    
    if (x + 1 < width)
    {
        // right
        offsetIndex = index + 1;
        offsetPixel = original[offsetIndex];  //lg01 - reduce error values from 7 to 6
        offsetPixel.Red = ToByte((offsetPixel.Red + ((redError * 6) >> 4)));
        offsetPixel.Green = ToByte((offsetPixel.Green + ((greenError * 6) >> 4)));
        offsetPixel.Blue = ToByte((offsetPixel.Blue + ((blueError * 6) >> 4)));
        original[offsetIndex] = offsetPixel;
    }
    
    if (y + 1 < height)
    {
        if (x - 1 > 0)
        {
            // left and down
            offsetIndex = index + width - 1;
            offsetPixel = original[offsetIndex];
            offsetPixel.Red = ToByte((offsetPixel.Red + ((redError * 3) >> 4)));
            offsetPixel.Green = ToByte((offsetPixel.Green + ((greenError * 3) >> 4)));
            offsetPixel.Blue = ToByte((offsetPixel.Blue + ((blueError * 3) >> 4)));
            original[offsetIndex] = offsetPixel;
        }
        
        
   
        // down
        offsetIndex = index + width;
        offsetPixel = original[offsetIndex];
        offsetPixel.Red = ToByte((offsetPixel.Red + ((redError * 5) >> 4)));
        offsetPixel.Green = ToByte((offsetPixel.Green + ((greenError * 5) >> 4)));
        offsetPixel.Blue = ToByte((offsetPixel.Blue + ((blueError * 5) >> 4)));
        original[offsetIndex] = offsetPixel;
    
        
        if (x + 1 < width)
        {
            // right and down
            offsetIndex = index + width + 1;
            offsetPixel = original[offsetIndex];
            offsetPixel.Red = ToByte((offsetPixel.Red + ((redError * 1) >> 4)));
            offsetPixel.Green = ToByte((offsetPixel.Green + ((greenError * 1) >> 4)));
            offsetPixel.Blue = ToByte((offsetPixel.Blue + ((blueError * 1) >> 4)));
            original[offsetIndex] = offsetPixel;
        }
    }
}

/*###############################################################*/
/*
 CODE for converting monochrome BMP files
 to compressed PCX files.
 */
/*###############################################################*/

// User defined fundemental data types.
#define BYTE unsigned char
#define CHAR char
#define WORD unsigned short
#define SHORT short
#define DWORD unsigned int
#define LONG int


// BMP file identifier.
BYTE bmpid[2];


// Type definition, BMP file header (12 bytes).
typedef struct {
    DWORD FileSize;
    WORD  Reserve1;
    WORD  Reserve2;
    DWORD Offset;
}BMP_FILE_HDR;

BMP_FILE_HDR BmpHdr;


// Type definition, DIB header version 1 (12 bytes).
typedef struct {
    DWORD Size;
    WORD  Width;
    WORD  Height;
    WORD  Planes;
    WORD  BitsPerPixel;
}BMP_DIB_HDR1;

// Version 1, DIB file header.
BMP_DIB_HDR1 DibHdr1;


// Type definition, DIB header versions 2 and above (108 bytes max).
typedef struct {
    DWORD Size;
    LONG  Width;
    LONG  Height;
    WORD  Planes;
    WORD  BitsPerPixel;
    DWORD Compression;
    DWORD ImageSize;
    LONG  HorzResolution;
    LONG  VertResolution;
    DWORD Colors;
    DWORD ColorsImportant;
    BYTE  Reserve[72];
}BMP_DIB_HDR;

// Version 2, DIB file header.
BMP_DIB_HDR DibHdr;


// Type definition, PCX file header (128 bytes).
typedef struct {
    BYTE  Identifier;        /* PCX Id Number (Always 0x0A) */
    BYTE  Version;           /* Version Number */
    BYTE  Encoding;          /* Encoding Format */
    BYTE  BitsPerPixel;      /* Bits per Pixel */
    WORD  XStart;            /* Left of image */
    WORD  YStart;            /* Top of Image */
    WORD  XEnd;              /* Right of Image */
    WORD  YEnd;              /* Bottom of image */
    WORD  HorzRes;           /* Horizontal Resolution */
    WORD  VertRes;           /* Vertical Resolution */
    BYTE  Palette[48];       /* 16-Color EGA Palette */
    BYTE  Reserved1;         /* Reserved (Always 0) */
    BYTE  NumBitPlanes;      /* Number of Bit Planes */
    WORD  BytesPerLine;      /* Bytes per Scan-line */
    WORD  PaletteType;       /* Palette Type */
    WORD  HorzScreenSize;    /* Horizontal Screen Size */
    WORD  VertScreenSize;    /* Vertical Screen Size */
    BYTE  Reserved2[54];     /* Reserved (Always 0) */
}PCX_FILE_HDR;

// PCX header.
PCX_FILE_HDR PcxHdr;

// PCX end of row bit mask.
BYTE Mask[] = {0x00, 0x80, 0xC0, 0xE0, 0xF0, 0xF8, 0xFC, 0xFE, 0xFF};



/*###############################################################*/
/*     */
/* Support Routines     */
/*     */
/*###############################################################*/

// Clear row buffer.
void ClearBuffer(BYTE *pBuffer, DWORD length)
{
    DWORD i;
    
    for(i=0; i<length; i++)
    *pBuffer++ = 0;
}


// Display text messages on console.
void DisplayMessage(CHAR *pMessage)
{
    printf("\r\n%s", pMessage);
}


/*###############################################################*/
/*     */
/* PCX Graphic Routines     */
/*     */
/*###############################################################*/

// Display PCX header.
void PCX_DisplayHeader(void)
{
    NSLog(@"PCX File Header");
    NSLog(@"Version    = %u", PcxHdr.Version);
    NSLog(@"Encoding   = %u", PcxHdr.Encoding);
    NSLog(@"Bits/Pixel = %u", PcxHdr.BitsPerPixel);
    NSLog(@"X Start    = %u", PcxHdr.XStart);
    NSLog(@"Y Start    = %u", PcxHdr.YStart);
    NSLog(@"X End      = %u", PcxHdr.XEnd);
    NSLog(@"Y End      = %u", PcxHdr.YEnd);
    NSLog(@"Horz Res   = %u", PcxHdr.HorzRes);
    NSLog(@"Vert Res   = %u", PcxHdr.VertRes);
    NSLog(@"Bits/Plane = %u", PcxHdr.NumBitPlanes);
    NSLog(@"Bytes/Line = %u", PcxHdr.BytesPerLine);
    NSLog(@"Palette ID = %u", PcxHdr.PaletteType);
    NSLog(@"Horz Size  = %u", PcxHdr.HorzScreenSize);
    NSLog(@"Vert Size  = %u", PcxHdr.VertScreenSize);
}

// Display a row of PCX graphic data.
void PCX_DisplayRow(BYTE *pRow, DWORD length)
{
    DWORD i;
    
    printf("\r\n");
    for(i=0; i<length; i++)
    printf("%02X ", *pRow++);
}

// Build PCX file header using information from BMP file's
// DIB header.
void PCX_BuildHeader(BMP_DIB_HDR *pDibHdr)
{
    DWORD i;
    
    PcxHdr.Identifier = 0x0A;
    PcxHdr.Version = 5;
    PcxHdr.Encoding = 1;
    PcxHdr.BitsPerPixel = pDibHdr->BitsPerPixel;
    PcxHdr.XStart = 0;
    PcxHdr.YStart = 0;
    PcxHdr.XEnd = pDibHdr->Width - 1;
    PcxHdr.YEnd = pDibHdr->Height - 1;
    PcxHdr.HorzRes = 0;
    PcxHdr.VertRes = 0;
    
    // Clear PCX color palette.
    for(i=0; i<48; i++)
    PcxHdr.Palette[i] = 0;
    
    // Set first two entries for black and white.
    for(i=3; i<6; i++)
    PcxHdr.Palette[i] = 0xFF;
    
    PcxHdr.Reserved1 = 0;
    PcxHdr.NumBitPlanes = 1;
    PcxHdr.BytesPerLine = (pDibHdr->Width + 7) / 8;
    PcxHdr.PaletteType = 1;
    PcxHdr.HorzScreenSize = 0;
    PcxHdr.VertScreenSize = 0;
    
    // Reserved bytes in header file, set to 0.
    for(i=0; i<54; i++)
    PcxHdr.Reserved2[i] = 0;
}



// Compress a row of BMP graphic data and save the new row as
// PCX graphic data.
DWORD PCX_CompressRow(BYTE *pPcx, BYTE *pBmp, DWORD length)
{
    DWORD index;
    BYTE control, data, runlen;
    
    control = 0xC0;
    data = 0;
    runlen = 0;
    index = 0;
    
    data = *pBmp++;
    runlen++;
    if (length < 8)  //lg01 - prevent negative
    {
        length = 8;
    }
    else
    {
    length = length - 8;
    }
    
    // While not last byte of BMP row, loop.
    while(length > 7)
    {
        // If next data byte equal current data byte, update RLE.
        if(data == *pBmp)
        {
            // Update run length count.
            runlen++;
            
            // If pcx row has exceeded maximum run length then
            // save this run and start new one.
            if(runlen >= 63)
            {
                // Save control byte and byte count.
                pPcx[index++] = control | runlen;
                
                // Save data byte, reset control byte and run length.
                pPcx[index++] = data;
                control = 0xC0;
                runlen = 0;
                
                //@@rp Move to next byte in BMP row.
                pBmp++;
                length = length - 8;
                //@@rp
                
                // Start new run, update data byte and run length.
                data = *pBmp;
                runlen++;
            }
        }
        else
        {
            // If run length greater than 1, save control byte.
            if(runlen > 1 || (data & 0xC0) == 0xC0)
            {
                pPcx[index++] = control | runlen;
            }
            
            // Save data byte, reset control byte and RLE count.
            pPcx[index++] = data;
            control = 0xC0;
            runlen = 0;
            
            // Start new run, update data byte and run length.
            data = *pBmp;
            runlen++;
        }
        
        // Update BMP row length and move to next byte.
        if (length < 8)  //lg01 - prevent negative
        {
            length = 8;
        }
        else
        {
        length = length - 8;
        }        pBmp++;
    }
    
    // If run length greater than 1, save control byte and data.
    if(runlen > 0)
    {
        //@@rp If run length greater than 1, save control byte.
        if(runlen > 1 || (data & 0xC0) == 0xC0)
        {
            pPcx[index++] = control | runlen;
        }
        //@@rp
        
        pPcx[index++] = data;
        
        // Reset control byte and RLE count.
        control = 0xC0;
        runlen = 0;
    }
    
    // If remaining length is greater than zero, save last byte
    // of BMP row in PCX row, mask off any unused bits at end of
    // row.
    if(length > 0)
    {
        // Load last byte from row buffer.
        data = *pBmp;
        runlen++;
        
        // If most significant bits of data set, save control byte.
        if((data & 0xC0) == 0xC0)
        {
            pPcx[index++] = control | runlen;
        }
        
        pPcx[index++] = (data & Mask[length]);
    }
    
    return index;
}

// Compress a row of BMP graphic data and save the new row as
// PCX graphic data.
DWORD PCX_CompressRowCount(BYTE *pBmp, DWORD length)
{
    DWORD index;
    BYTE control, data, runlen;
    
    control = 0xC0;
    data = 0;
    runlen = 0;
    index = 0;
    
    data = *pBmp++;
    runlen++;
    if (length < 8)  //lg01 - prevent negative
    {
        length = 8;
    }
    else
    {
    length = length - 8;
    }

    // While not last byte of BMP row, loop.
    while(length > 7)
    {
        // If next data byte equal current data byte, update RLE.
        if(data == *pBmp)
        {
            // Update run length count.
            runlen++;
            
            // If pcx row has exceeded maximum run length then
            // save this run and start new one.
            if(runlen >= 63)
            {
                // Save control byte and byte count.
                index++;
                
                // Save data byte, reset control byte and run length.
                index++;
                control = 0xC0;
                runlen = 0;
                
                //@@rp Move to next byte in BMP row.
                pBmp++;
                if (length < 8)  //lg01 - prevent negative
                {
                    length = 8;
                }
                else
                {
                length = length - 8;
                }                //@@rp
                
                // Start new run, update data byte and run length.
                data = *pBmp;
                runlen++;
            }
        }
        else
        {
            // If run length greater than 1, save control byte.
            if(runlen > 1 || (data & 0xC0) == 0xC0)
            {
                index++;
            }
            
            // Save data byte, reset control byte and RLE count.
            index++;
            control = 0xC0;
            runlen = 0;
            
            // Start new run, update data byte and run length.
            data = *pBmp;
            runlen++;
        }
        
        // Update BMP row length and move to next byte.
        if (length < 8)  //lg01 - prevent negative
        {
            length = 8;
        }
        else
        {
        length = length - 8;
        }
        pBmp++;
    }
    
    // If run length greater than 1, save control byte and data.
    if(runlen > 0)
    {
        //@@rp If run length greater than 1, save control byte.
        if(runlen > 1 || (data & 0xC0) == 0xC0)
        {
            index++;
        }
        //@@rp
        
        index++;
        
        // Reset control byte and RLE count.
        control = 0xC0;
        runlen = 0;
    }
    
    // If remaining length is greater than zero, save last byte
    // of BMP row in PCX row, mask off any unused bits at end of
    // row.
    if(length > 0)
    {
        // Load last byte from row buffer.
        data = *pBmp;
        runlen++;
        
        // If most significant bits of data set, save control byte.
        if((data & 0xC0) == 0xC0)
        {
            index++;
        }
        
        index++;
    }
    
    return index;
}


/*###############################################################*/
/*     */
/* BMP Graphic Routines     */
/*     */
/*###############################################################*/

// Display BMP file header.
void BMP_DisplayHeader(void)
{
    printf("\nBMP File Header");
    printf("\nFile Size   = %u", BmpHdr.FileSize);
    printf("\nFile Offset = %u", BmpHdr.Offset);
    printf("\n");
}


// Display DIB header.
void DIB_DisplayHeader(void)
{
    DWORD rowlen;
    
    // Add row length to header info.
    rowlen = ((DibHdr.Width + 31) / 32) * 4;
    
    printf("\nDIB Header");
    printf("\nHeader Size  = %u", DibHdr.Size);
    printf("\nImage Width  = %u", DibHdr.Width);
    printf("\nImage Height = %u", DibHdr.Height);
    printf("\nImage Planes = %u", DibHdr.Planes);
    printf("\nBits / Pixel = %u", DibHdr.BitsPerPixel);
    printf("\nCompression  = %u", DibHdr.Compression);
    printf("\nImage Size   = %u", DibHdr.ImageSize);
    printf("\nH Resolution = %u", DibHdr.HorzResolution);
    printf("\nV Resolution = %u", DibHdr.VertResolution);
    printf("\nColors       = %u", DibHdr.Colors);
    printf("\nImportant    = %u", DibHdr.ColorsImportant);
    
    // This is not part of header but is include due to importance!
    printf("\r\nRow Length   = %u", rowlen);
    
    printf("\r\n");
}


// Display row of BMP graphic data.
void BMP_DisplayRow(BYTE *pRow, DWORD length)
{
    DWORD i;
    
    printf("\r\n");
    for(i=0; i<length; i++)
    {
        printf("%02X ", *pRow++);
    }
}

const BYTE CRC7_POLY = 0x91;

BYTE getCRC(BYTE* message, DWORD length)
{
    DWORD i, j;
    BYTE crc = 0;
    BYTE* p;
    BYTE b;
    
    for (p = message, i = 0; i < length; i++, p++)
    {
        b = *p;
        crc ^= b;
        for (j = 0; j < 8; j++)
        {
            if (crc & 1)
                crc ^= CRC7_POLY;
            crc >>= 1;
        }
    }
    return crc;
}

/*###############################################################*/
/*     */
/* BmpToPcx     */
/*     */
/*###############################################################*/
bool BmpToPcx(NSData *myData)
{
    //move bitmap file to array
    const BYTE *pBmpBuffer = (const BYTE*)[myData bytes];
    BYTE *pPcxBuffer = NULL;
    BYTE* pBuffer;
    size_t index;
    DWORD size, bmprowlen, i, pcxrowlen;
    DWORD myDataSize = 0;
    DWORD pcxSize1 = 0;
    DWORD pcxSize2 = 0;
    BYTE* pBmp = NULL;
    BYTE* pPcx = NULL;
    @try {
        //printf("BmpToPcx: Data Length = %u\n", (DWORD)myData.length);
        // Read BMP file identifier, two bytes long.
        pBmp = (BYTE*)pBmpBuffer;
        memcpy(&bmpid[0], pBmp, 2);
        pBmp += 2;
        
        myDataSize = (DWORD)myData.length;
        //NSLog(@"CRC = 0x%02x, length = %u", getCRC(pBmpBuffer, myDataSize), myDataSize);
        
        // Error, If bad file identifier display message and exit program.
        if (bmpid[0] != 'B' || bmpid[1] != 'M')
        {
            DisplayMessage("BMP file has bad file identifier.");
            return false;
        }
        
        // Read BMP file header.
        memcpy(&BmpHdr, pBmp, sizeof(BMP_FILE_HDR));
        pBmp += sizeof(BMP_FILE_HDR);
        
        // Assign pointer to version 1 of DIB header.
        pBuffer = (BYTE*)&DibHdr1;
        
        // Determine size of first field (header size).
        size = sizeof(DibHdr1.Size);
        
        // Read size of DIB header.
        memcpy(pBuffer, pBmp, size);
        pBmp += size;
        
        // If DIB header is old style then read remaining header into
        // original data structure and transfer old header to new header.
        if (DibHdr1.Size <= sizeof(BMP_DIB_HDR1))
        {
            // Adjust size of remaining header and pointer to file header.
            pBuffer = pBuffer + size;
            size = DibHdr1.Size - size;
            
            // Read remaining header.
            memcpy(pBuffer, pBmp, size);
            
            // Transfer old header values to new header.
            DibHdr.Size = DibHdr1.Size;
            DibHdr.Width = DibHdr1.Width;
            DibHdr.Height = DibHdr1.Height;
            DibHdr.Planes = DibHdr1.Planes;
            DibHdr.BitsPerPixel = DibHdr1.BitsPerPixel;
        }
        else // New Header
        {
            // Transfer header size from original header to new header.
            DibHdr.Size = DibHdr1.Size;
            
            // Point to new header and adjust pointer for remaining data.
            pBuffer = (BYTE*)&DibHdr;
            pBuffer = pBuffer + size;
            
            // Calculate size of remaining header.
            size = DibHdr.Size - size;
            
            // Read remaining data for DIB header.
            memcpy(pBuffer, pBmp, size);
            pBmp += size;
        }
        
        // Check for illegal file format (color or gray scale).
        if (DibHdr.Planes > 1 || DibHdr.BitsPerPixel > 1)
        {
            DisplayMessage("BMP file not monochrome.");
            return false;
        }
        
        // Move file pointer to start of bitmap data.
        pBmp = (BYTE*)(pBmpBuffer + BmpHdr.Offset);
        
        DWORD dwSize = BmpHdr.Offset;
        // printf("\nCRC Header = 0x%02x, length = %u", getCRC(pBmpBuffer, dwSize), dwSize);
        dwSize = myDataSize - BmpHdr.Offset;
        // printf("\nImage Data Size = %u", dwSize);
        //printf("\nCRC Image  = 0x%02x, length = %u", getCRC(pBmp, dwSize), dwSize);
        
        // Display BMP file header.
        // BMP_DisplayHeader();
        
        // Display DIB header.
        // DIB_DisplayHeader();
        
        // Calculate number of bytes in BMP row.
        bmprowlen = ((DibHdr.Width + 31) / 32) * 4;
        
        
        // Calculate starting position of last row in BMP file.
        // Note: BMP rows are stored in reverse order.
        index = (bmprowlen * (DibHdr.Height - 1)) + BmpHdr.Offset;
        pBmp = (BYTE*)pBmpBuffer + index;
        
        pcxSize1 = sizeof(PCX_FILE_HDR);
        pcxSize2 = sizeof(PCX_FILE_HDR);
        //("sizeof(PCX_FILE_HDR) = %u, %u\n", pcxSize1, pcxSize2);
        
        // Read a row of graphic's data from the BMP file, compress the row using
        // RLE and save compressed data in PCX row buffer. Write row of compressed
        // data to PCX file.
        for (i = 0; i < (unsigned int)DibHdr.Height; i++)
        {
            // Display BMP Input row.
            // BMP_DisplayRow(pBmp, bmprowlen); //@test
            
            // Get Size of RLE Compress graphic data.
            pcxrowlen = PCX_CompressRowCount(pBmp, DibHdr.Width);
            pcxSize1 += pcxrowlen;
            pBmp -= bmprowlen;
        }
        pPcxBuffer = malloc(pcxSize1);
        if (NULL == pPcxBuffer)
        {
            printf("Error, pPcxBuffer is NULL\n");
            return false;
        }
        //printf("BmpToPcx: pPcxBuffer = %p, pcxSize1 = %u\n", pPcxBuffer, pcxSize1);
        pPcx = pPcxBuffer;
        PCX_BuildHeader(&DibHdr);
        // PCX_DisplayHeader();
        memcpy(pPcx, &PcxHdr, sizeof(PCX_FILE_HDR));
        pPcx += sizeof(PCX_FILE_HDR);
        
        // Calculate starting position of last row in BMP file.
        // Note: BMP rows are stored in reverse order.
        index = (bmprowlen * (DibHdr.Height - 1)) + BmpHdr.Offset;
        pBmp = (BYTE*)pBmpBuffer + index;
        
        // Read a row of graphic's data from the BMP file, compress the row using
        // RLE and save compressed data in PCX row buffer. Write row of compressed
        // data to PCX file.
        for (i = 0; i < (unsigned int)DibHdr.Height; i++)
        {
            // Display BMP Input row.
            // BMP_DisplayRow(pBmp, bmprowlen); //@test
            
            // Compress graphic data and save in PCX row buffer.
            pcxrowlen = PCX_CompressRow(pPcx, pBmp, DibHdr.Width);
            pcxSize2 += pcxrowlen;
            pPcx += pcxrowlen;
            // Set pBmp pointer to previous bitmap row
            pBmp -= bmprowlen;
        }
        //printf("BmpToPcx: pPcxBuffer = %p, pcxSize2 = %u\n", pPcxBuffer, pcxSize2);
        if (pcxSize1 != pcxSize2)
        {
            free(pPcxBuffer);
            pPcxBuffer = NULL;
            printf("Error, pcxSize1 != pcxSize2\n");
            return false;
        }
        
        NSString *command1 = @"";
        NSString *command2 = @"";
        
        if (0 == curRowPos || 0 == curColPos)
        {
            //if logo number does not exist then print the image else download graphic as a logo
            if([LogoNumber isEqual:@""])
            {
                //complete header command to include path, orientation, no repeat and original row column position
                command1 = [NSString stringWithFormat:@"<pcx><G%u>",pcxSize2];  //lg02 - use old resolution field to scale image
                if (PrinterResolution == 2)  //lg02 - double if scaled
                {
                    command1 = [NSString stringWithFormat:@"<HW2,2><pcx><G%u>", pcxSize2];
                }
                if (PrinterResolution == 3)  //lg02 - triple
                {
                    command1 = [NSString stringWithFormat:@"<HW3,3><pcx><G%u>", pcxSize2];
                }
             
                SendString(command1);
                SendData(pPcxBuffer, pcxSize2);
            }
            else        //download graphic as logo00
            {
                int logoNum = [LogoNumber intValue];
                command1 = [NSString stringWithFormat:@"%C<LD%d><pcx><G%u>", 0x001b, logoNum, pcxSize2];
                SendString(command1);
                SendData(pPcxBuffer, pcxSize2);
                command2 = [NSString stringWithFormat:@"%C", 0x001b];
                SendString(command2);
            }
        }
        else
        {
            //if logo number does not exist then print the image else download graphic as a logo
            if([LogoNumber isEqual:@""])
            {
                //complete header command to include path, orientation, no repeat and original row column position
                command1 = [NSString stringWithFormat:@"<SP%d,%d><pcx><G%u>", curRowPos, curColPos, pcxSize2];  //lg02 - use old resolution field to scale image
                if (PrinterResolution == 2)  //lg02 - double if scaled
                {
                    command1 = [NSString stringWithFormat:@"<SP%d,%d><HW2,2><pcx><G%u>", curRowPos, curColPos, pcxSize2];
                }
                if (PrinterResolution == 3)  //lg02 - triple
                {
                    command1 = [NSString stringWithFormat:@"<SP%d,%d><HW3,3><pcx><G%u>",curRowPos, curColPos, pcxSize2];
                }
             
                SendString(command1);
                SendData(pPcxBuffer, pcxSize2);
            }
            else        //download graphic as logo00
            {
                int logoNum = [LogoNumber intValue];
                command1 = [NSString stringWithFormat:@"<LD%d><SP%d,%d>%C<pcx><G%u>", logoNum, curRowPos, curColPos, 0x001b, pcxSize2];
                SendString(command1);
                SendData(pPcxBuffer, pcxSize2);
                command2 = [NSString stringWithFormat:@"%C", 0x001b];
                SendString(command2);
            }
        }
            
        // printf("pPcxBuffer = %p\n", pPcxBuffer);
        if (NULL != pPcxBuffer)
        {
            free(pPcxBuffer);
            pPcxBuffer = NULL;
        }
    }
    @catch(NSException *exception)
    {
        if (NULL !=pPcxBuffer)
        {
            free(pPcxBuffer);
        }
        NSLog(@"Exception: %@", exception.reason);
    }
    
    // CFTimeInterval elapsedTime = CACurrentMediaTime() - startTime;
    // NSLog(@"Elapsed Time: %f", elapsedTime);
    
    // Exit program.
    return true;
}

//When rotating into portrait mode, there is a little positioning error with default setting
//The dot array below is used to reduce the positioning error.  Some of the numbers are
//approximates and can use some fine tuning if used.
void InitDotArray()
{
    //concert stock 2"
    portrait_dot_array[0][0] = 384;     //200 DPI
    portrait_dot_array[0][1] = 576;     //300 DPI
    portrait_dot_array[0][2] = 1184;    //600 DPI
    //cinema stock 3.25"
    portrait_dot_array[1][0] = 640;     //200 DPI
    portrait_dot_array[1][1] = 960;     //300 DPI
    portrait_dot_array[1][2] = 1920;    //600 DPI approximate
    //credit card stock 2.125"
    portrait_dot_array[2][0] = 416;     //200 DPI
    portrait_dot_array[2][1] = 608;     //300 DPI
    portrait_dot_array[2][2] = 1232;    //600 DPI
    //receipt stock 3.25"
    portrait_dot_array[3][0] = 640;     //200 DPI
    portrait_dot_array[3][1] = 960;     //300 DPI
    portrait_dot_array[3][2] = 1920;    //600 DPI approximate
    //ski stock 3.25"
    portrait_dot_array[4][0] = 640;     //200 DPI
    portrait_dot_array[4][1] = 960;     //300 DPI
    portrait_dot_array[4][2] = 1920;    //600 DPI approximate
    //4x6 stock
    portrait_dot_array[5][0] = 816;     //200 DPI
    portrait_dot_array[5][1] = 1216;    //300 DPI
    portrait_dot_array[5][2] = 2360;    //600 DPI approximate
    //Wristband 1 stock 1"
    portrait_dot_array[6][0] = 190;     //200 DPI approximate
    portrait_dot_array[6][1] = 290;     //300 DPI approximate
    portrait_dot_array[6][2] = 590;     //600 DPI approximate
    //Wristband 2 stock 1.328"
    portrait_dot_array[7][0] = 250;     //200 DPI approximate
    portrait_dot_array[7][1] = 380;     //300 DPI approximate
    portrait_dot_array[7][2] = 780;     //600 DPI approximate
    //letter/custom stock
    portrait_dot_array[8][0] = 1700;    //200 DPI approximate
    portrait_dot_array[8][1] = 2580;    //300 DPI approximate
    portrait_dot_array[8][2] = 5000;    //600 DPI approximate
    
    set_portrait_dots();
    
}

//read portrait dots width from array
void set_portrait_dots()
{
    if ([PrinterOrientation isEqual: @"<PM>"])
    {
        //retrieve dots per width (lptix) in portrait mode
        switch (PrinterResolution)
        {
            case 200:
                portrait_dots = portrait_dot_array[size_indicator][0];
                break;
            case 300:
                portrait_dots = portrait_dot_array[size_indicator][1];
                break;
            case 600:
                portrait_dots = portrait_dot_array[size_indicator][2];
                break;
                
        }
    }
    else
    {
        portrait_dots = 0;
    }
}

//To clear User space memory, send an Escape C, to the printer
void ClearMemory()
{
    NSString *command = @"";
    command = [NSString stringWithFormat:@"%@%C",command,0x001b];           //include escape control character
    command = [NSString stringWithFormat:@"%@%@",command,@"c"];             //Add lower case C to the string
    SendString(command);
    
}

bool SendFileImage(UIImage *image, int row, int column)
{
    bool status = true;
    LogoNumber = @"";
    if (CUR_POS != row)
    {
        curRowPos = row;
    }
    if (CUR_POS != column)
    {
        curColPos = column;
    }
    int BMP_WIDTH;
    int BMP_HEIGHT;
   
    // NSLog(@"SendFileImage");
    
    //Convert pixels per inch size to Dots per inch size
    BMP_WIDTH = (int)((image.size.width/PPI) * PrinterResolution);
    BMP_HEIGHT = (int)((image.size.height/PPI) * PrinterResolution);
    
    //NSLog(@"PPI Width  = %f",result.size.width);
    //NSLog(@"PPI Height = %f",result.size.height);
    //NSLog(@"DPI Width  = %d",BMP_WIDTH);
    //NSLog(@"DPI Height = %d",BMP_HEIGHT);
 
    //Print the resulting image
    PrintBMP(BMP_WIDTH, BMP_HEIGHT, image);
    
    return(status);
}

/*
 
 FOUNDATION_EXPORT NSString *NSUserName(void);
 FOUNDATION_EXPORT NSString *NSFullUserName(void);

 FOUNDATION_EXPORT NSString *NSHomeDirectory(void);
 FOUNDATION_EXPORT NSString * _Nullable NSHomeDirectoryForUser(NSString * _Nullable userName);

 FOUNDATION_EXPORT NSString *NSTemporaryDirectory(void);

 FOUNDATION_EXPORT NSString *NSOpenStepRootDirectory(void);

 typedef NS_ENUM(NSUInteger, NSSearchPathDirectory) {
     NSApplicationDirectory = 1,             // supported applications (Applications)
     NSDemoApplicationDirectory,             // unsupported applications, demonstration versions (Demos)
     NSDeveloperApplicationDirectory,        // developer applications (Developer/Applications). DEPRECATED - there is no one single Developer directory.
     NSAdminApplicationDirectory,            // system and network administration applications (Administration)
     NSLibraryDirectory,                     // various documentation, support, and configuration files, resources (Library)
     NSDeveloperDirectory,                   // developer resources (Developer) DEPRECATED - there is no one single Developer directory.
     NSUserDirectory,                        // user home directories (Users)
     NSDocumentationDirectory,               // documentation (Documentation)
     NSDocumentDirectory,                    // documents (Documents)
     NSCoreServiceDirectory,                 // location of CoreServices directory (System/Library/CoreServices)
     NSAutosavedInformationDirectory API_AVAILABLE(macos(10.6), ios(4.0), watchos(2.0), tvos(9.0)) = 11,   // location of autosaved documents (Documents/Autosaved)
     NSDesktopDirectory = 12,                // location of user's desktop
     NSCachesDirectory = 13,                 // location of discardable cache files (Library/Caches)
     NSApplicationSupportDirectory = 14,     // location of application support files (plug-ins, etc) (Library/Application Support)
     NSDownloadsDirectory API_AVAILABLE(macos(10.5), ios(2.0), watchos(2.0), tvos(9.0)) = 15,              // location of the user's "Downloads" directory
     NSInputMethodsDirectory API_AVAILABLE(macos(10.6), ios(4.0), watchos(2.0), tvos(9.0)) = 16,           // input methods (Library/Input Methods)
     NSMoviesDirectory API_AVAILABLE(macos(10.6), ios(4.0), watchos(2.0), tvos(9.0)) = 17,                 // location of user's Movies directory (~/Movies)
     NSMusicDirectory API_AVAILABLE(macos(10.6), ios(4.0), watchos(2.0), tvos(9.0)) = 18,                  // location of user's Music directory (~/Music)
     NSPicturesDirectory API_AVAILABLE(macos(10.6), ios(4.0), watchos(2.0), tvos(9.0)) = 19,               // location of user's Pictures directory (~/Pictures)
     NSPrinterDescriptionDirectory API_AVAILABLE(macos(10.6), ios(4.0), watchos(2.0), tvos(9.0)) = 20,     // location of system's PPDs directory (Library/Printers/PPDs)
     NSSharedPublicDirectory API_AVAILABLE(macos(10.6), ios(4.0), watchos(2.0), tvos(9.0)) = 21,           // location of user's Public sharing directory (~/Public)
     NSPreferencePanesDirectory API_AVAILABLE(macos(10.6), ios(4.0), watchos(2.0), tvos(9.0)) = 22,        // location of the PreferencePanes directory for use with System Preferences (Library/PreferencePanes)
     NSApplicationScriptsDirectory API_AVAILABLE(macos(10.8)) API_UNAVAILABLE(ios, watchos, tvos) = 23,      // location of the user scripts folder for the calling application (~/Library/Application Scripts/code-signing-id)
     NSItemReplacementDirectory API_AVAILABLE(macos(10.6), ios(4.0), watchos(2.0), tvos(9.0)) = 99,        // For use with NSFileManager's URLForDirectory:inDomain:appropriateForURL:create:error:
     NSAllApplicationsDirectory = 100,       // all directories where applications can occur
     NSAllLibrariesDirectory = 101,          // all directories where resources can occur
     NSTrashDirectory API_AVAILABLE(macos(10.8), ios(11.0)) API_UNAVAILABLE(watchos, tvos) = 102             // location of Trash directory

 };

 typedef NS_OPTIONS(NSUInteger, NSSearchPathDomainMask) {
     NSUserDomainMask = 1,       // user's home directory --- place to install user's personal items (~)
     NSLocalDomainMask = 2,      // local to the current machine --- place to install items available to everyone on this machine (/Library)
     NSNetworkDomainMask = 4,    // publically available location in the local area network --- place to install items available on the network (/Network)
     NSSystemDomainMask = 8,     // provided by Apple, unmodifiable (/System)
     NSAllDomainsMask = 0x0ffff  // all domains: all of the above and future items
 };

 FOUNDATION_EXPORT NSArray<NSString *> *NSSearchPathForDirectoriesInDomains(NSSearchPathDirectory directory, NSSearchPathDomainMask domainMask, BOOL expandTilde);

 
 
 
 */

#define STRINGIFY2( x) #x
#define STRINGIFY(x) STRINGIFY2(x)
// #define PASTE2( a, b) a##b
// #define PASTE( a, b) PASTE2( a, b)
// #define PRINTTHIS(text) \
//    NSLog(PASTE( @, STRINGIFY(text)));
 
// #define PrintDirectory(NS_ENUM, NS_ENUM


typedef struct
{
    char *disp;
    NSUInteger value;
} SEARCH_ARRAY;

SEARCH_ARRAY SearchArray[] =
{
    {"NSApplicationDirectory", NSApplicationDirectory},
    {"NSDemoApplicationDirectory", NSDemoApplicationDirectory},
    {"NSDeveloperApplicationDirectory", NSDeveloperApplicationDirectory},
    {"NSAdminApplicationDirectory", NSAdminApplicationDirectory},
    {"NSLibraryDirectory", NSLibraryDirectory},
    {"NSDeveloperDirectory", NSDeveloperDirectory},
    {"NSUserDirectory", NSUserDirectory},
    {"NSDocumentationDirectory", NSDocumentationDirectory},
    {"NSDocumentDirectory", NSDocumentDirectory},
    {"NSCoreServiceDirectory", NSCoreServiceDirectory},
    {"NSAutosavedInformationDirectory", NSAutosavedInformationDirectory},
    {"NSDesktopDirectory", NSDesktopDirectory},
    {"NSCachesDirectory", NSCachesDirectory},
    {"NSApplicationSupportDirectory", NSApplicationSupportDirectory},
    {"NSDownloadsDirectory", NSDownloadsDirectory},
    {"NSInputMethodsDirectory", NSInputMethodsDirectory},
    {"NSMoviesDirectory", NSMoviesDirectory},
    {"NSMusicDirectory", NSMusicDirectory},
    {"NSPicturesDirectory", NSPicturesDirectory},
    {"NSPrinterDescriptionDirectory", NSPrinterDescriptionDirectory},
    {"NSSharedPublicDirectory", NSSharedPublicDirectory},
    {"NSPreferencePanesDirectory", NSPreferencePanesDirectory},
    // {"NSApplicationScriptsDirectory", NSApplicationScriptsDirectory},
    {"NSItemReplacementDirectory", NSItemReplacementDirectory},
    {"NSAllApplicationsDirectory", NSAllApplicationsDirectory},
    {"NSAllLibrariesDirectory", NSAllLibrariesDirectory},
    {"NSTrashDirectory", NSTrashDirectory}
};
const NSUInteger SearchArrayCount = sizeof(SearchArray) / sizeof(SEARCH_ARRAY);

typedef struct
{
    char *disp;
    NSUInteger value;
} DOMAIN_ARRAY;

DOMAIN_ARRAY DomainArray[] =
{
    {"NSUserDomainMask", NSUserDomainMask},
    {"NSLocalDomainMask", NSLocalDomainMask},
    {"NSNetworkDomainMask", NSNetworkDomainMask},
    {"NSSystemDomainMask", NSSystemDomainMask},
    //{"NSAllDomainsMask", NSAllDomainsMask}
};
const NSUInteger DomainArrayCount = sizeof(DomainArray) / sizeof(DOMAIN_ARRAY);

void ListUserNameAndDirectories(void)
{
    NSArray *paths = nil;
    printf("\n\n");
    printf("NSUserName() = %s\n", [NSUserName() UTF8String]);
    printf("NSFullUserName() = %s\n", [NSFullUserName() UTF8String]);
    printf("NSHomeDirectory() = %s\n", [NSHomeDirectory() UTF8String]);
    printf("NSHomeDirectoryForUser(NSUserName()) = %s\n", [NSHomeDirectoryForUser(NSUserName()) UTF8String]);
    printf("NSTemporaryDirectory() = %s\n", [NSTemporaryDirectory() UTF8String]);
      
    printf("\n");
    // printf("SearchArrayCount = %lu\n", SearchArrayCount);
    // printf("DomainArrayCount = %lu\n", DomainArrayCount);
    
    for (NSUInteger i = 0; i < SearchArrayCount; i++)
    {
        for (NSUInteger j = 0; j < DomainArrayCount; j++)
        {
            @try {
                paths = NSSearchPathForDirectoriesInDomains(SearchArray[i].value, DomainArray[j].value, true);
                for (int k = 0; k < [paths count]; k++)
                {
                    NSString *val = [paths objectAtIndex:k];
                    printf("NSSearchPathForDirectoriesInDomains(%s, %s)[%d] =\n   %s\n", SearchArray[i].disp, DomainArray[j].disp, k, [val UTF8String]);
                }
            }
            @catch (NSException *exception)
            {
                printf("Exception: %s\n", [exception.name UTF8String]);
                printf("Reason: %s\n", [exception.reason UTF8String]);
            }
            paths = nil;
        }
    }
}

NSString *MakeSubdirInHomeDirectory(NSString *subDir)
{
    NSString *fullPathname = [NSHomeDirectory() stringByAppendingPathComponent:[NSString stringWithFormat:@"/%@", subDir]];
    return fullPathname;
}

bool CreateDirectory(NSString *pathname)
{
    NSFileManager *fm = [NSFileManager defaultManager];
    if ([fm fileExistsAtPath:pathname])
    {
        BOOL isDir = NO;
        if ([fm fileExistsAtPath:pathname isDirectory:&isDir] && isDir)
        {
            return true;
        }
        NSLog(@"Failed to Create Directory: %@", pathname);
        NSLog(@"It already exists but it is not a Directory");
        return false;
    }
    NSError *error = nil;
    if (![fm createDirectoryAtPath:pathname withIntermediateDirectories:YES attributes:nil error:&error])
    {
        NSLog(@"Failed to Create Directory: %@", pathname);
        NSLog(@"Error: %@", error);
        return false;
    }
    return true;
}

NSData *ReadFile(NSString *pathname)
{
    NSData *data = nil;
    data = [NSData dataWithContentsOfFile:pathname];
    if (nil == data)
    {
        NSLog(@"Could not Read File: %@", pathname);
    }
    return data;
}

bool WriteDataToFile(NSString *pathname, NSData *data)
{
    NSError *error = nil;
    [data writeToFile:pathname options:NSDataWritingAtomic error:&error];
    if (nil != error)
    {
        NSLog(@"Write returned error: %@", [error localizedDescription]);
        return false;
    }
    return true;
}

@end
 
